@@ -102,6 +102,8 @@ static inline void	   object_set_property		(GObject        *object,
 							 GParamSpec     *pspec,
 							 const GValue   *value,
 							 GObjectNotifyQueue *nqueue);
+static guint               object_floating_flag_handler (GObject        *object,
+                                                         gint            job);
 
 static void object_interface_check_properties           (gpointer        func_data,
 							 gpointer        g_iface);
@@ -114,8 +116,9 @@ static GQuark	            quark_toggle_refs = 0;
 static GParamSpecPool      *pspec_pool = NULL;
 static GObjectNotifyContext property_notify_context = { 0, };
 static gulong	            gobject_signals[LAST_SIGNAL] = { 0, };
-G_LOCK_DEFINE_STATIC (construct_objects_lock);
-static GSList *construct_objects = NULL;
+static guint (*floating_flag_handler) (GObject*, gint) = object_floating_flag_handler;
+G_LOCK_DEFINE_STATIC (construction_mutex);
+static GSList *construction_objects = NULL;
 
 /* --- functions --- */
 #ifdef	G_ENABLE_DEBUG
@@ -473,16 +476,14 @@ g_object_init (GObject *object)
 {
   object->ref_count = 1;
   g_datalist_init (&object->qdata);
-  g_object_force_floating (object);
   
   /* freeze object's notification queue, g_object_newv() preserves pairedness */
   g_object_notify_queue_freeze (object, &property_notify_context);
+  /* enter construction list for notify_queue_thaw() and to allow construct-only properties */
+  G_LOCK (construction_mutex);
+  construction_objects = g_slist_prepend (construction_objects, object);
+  G_UNLOCK (construction_mutex);
 
-  /* allow construct-only properties to be set */
-  G_LOCK (construct_objects_lock);
-  construct_objects = g_slist_prepend (construct_objects, object);
-  G_UNLOCK (construct_objects_lock);
-  
 #ifdef	G_ENABLE_DEBUG
   IF_DEBUG (OBJECTS)
     {
@@ -795,12 +796,34 @@ g_object_new (GType	   object_type,
 }
 
 static gboolean
-object_in_construction (GObject *object)
+slist_maybe_remove (GSList       **slist,
+                    gconstpointer  data)
+{
+  GSList *last = NULL, *node = *slist;
+  while (node)
+    {
+      if (node->data == data)
+        {
+          if (last)
+            last->next = node->next;
+          else
+            *slist = node->next;
+          g_slist_free_1 (node);
+          return TRUE;
+        }
+      last = node;
+      node = last->next;
+    }
+  return FALSE;
+}
+
+static inline gboolean
+object_in_construction_list (GObject *object)
 {
   gboolean in_construction;
-  G_LOCK (construct_objects_lock);
-  in_construction = g_slist_find (construct_objects, object) != NULL;
-  G_UNLOCK (construct_objects_lock);
+  G_LOCK (construction_mutex);
+  in_construction = g_slist_find (construction_objects, object) != NULL;
+  G_UNLOCK (construction_mutex);
   return in_construction;
 }
 
@@ -810,13 +833,14 @@ g_object_newv (GType       object_type,
 	       GParameter *parameters)
 {
   GObjectConstructParam *cparams, *oparams;
-  GObjectNotifyQueue *nqueue;
+  GObjectNotifyQueue *nqueue = NULL; /* shouldn't be initialized, just to silence compiler */
   GObject *object;
   GObjectClass *class, *unref_class = NULL;
   GSList *slist;
   guint n_total_cparams = 0, n_cparams = 0, n_oparams = 0, n_cvalues;
   GValue *cvalues;
   GList *clist = NULL;
+  gboolean newly_constructed;
   guint i;
 
   g_return_val_if_fail (G_TYPE_IS_OBJECT (object_type), NULL);
@@ -908,31 +932,32 @@ g_object_newv (GType       object_type,
 
   /* construct object from construction parameters */
   object = class->constructor (object_type, n_total_cparams, cparams);
-  G_LOCK (construct_objects_lock);
-  construct_objects = g_slist_remove (construct_objects, object);
-  G_UNLOCK (construct_objects_lock);
-
   /* free construction values */
   g_free (cparams);
   while (n_cvalues--)
     g_value_unset (cvalues + n_cvalues);
   g_free (cvalues);
-  
-  /* release g_object_init() notification queue freeze_count */
-  nqueue = g_object_notify_queue_freeze (object, &property_notify_context);
-  g_object_notify_queue_thaw (object, nqueue);
-  
+
+  /* adjust freeze_count according to g_object_init() and remaining properties */
+  G_LOCK (construction_mutex);
+  newly_constructed = slist_maybe_remove (&construction_objects, object);
+  G_UNLOCK (construction_mutex);
+  if (newly_constructed || n_oparams)
+    nqueue = g_object_notify_queue_freeze (object, &property_notify_context);
+  if (newly_constructed)
+    g_object_notify_queue_thaw (object, nqueue);
+
   /* set remaining properties */
   for (i = 0; i < n_oparams; i++)
     object_set_property (object, oparams[i].pspec, oparams[i].value, nqueue);
   g_free (oparams);
 
+  /* release our own freeze count and handle notifications */
+  if (newly_constructed || n_oparams)
+    g_object_notify_queue_thaw (object, nqueue);
   if (unref_class)
     g_type_class_unref (unref_class);
 
-  /* release our own freeze count and handle notifications */
-  g_object_notify_queue_thaw (object, nqueue);
-  
   return object;
 }
 
@@ -1076,7 +1101,7 @@ g_object_set_valist (GObject	 *object,
 		     G_OBJECT_TYPE_NAME (object));
 	  break;
 	}
-      if ((pspec->flags & G_PARAM_CONSTRUCT_ONLY) && !object_in_construction (object))
+      if ((pspec->flags & G_PARAM_CONSTRUCT_ONLY) && !object_in_construction_list (object))
         {
           g_warning ("%s: construct property \"%s\" for object `%s' can't be set after construction",
                      G_STRFUNC, pspec->name, G_OBJECT_TYPE_NAME (object));
@@ -1224,7 +1249,7 @@ g_object_set_property (GObject	    *object,
                G_STRFUNC,
                pspec->name,
                G_OBJECT_TYPE_NAME (object));
-  else if ((pspec->flags & G_PARAM_CONSTRUCT_ONLY) && !object_in_construction (object))
+  else if ((pspec->flags & G_PARAM_CONSTRUCT_ONLY) && !object_in_construction_list (object))
     g_warning ("%s: construct property \"%s\" for object `%s' can't be set after construction",
                G_STRFUNC, pspec->name, G_OBJECT_TYPE_NAME (object));
   else
@@ -1520,29 +1545,48 @@ g_object_remove_weak_pointer (GObject  *object,
                        weak_pointer_location);
 }
 
+static guint
+object_floating_flag_handler (GObject        *object,
+                              gint            job)
+{
+  switch (job)
+    {
+      gpointer oldvalue;
+    case +1:    /* force floating if possible */
+      do
+        oldvalue = g_atomic_pointer_get (&object->qdata);
+      while (!g_atomic_pointer_compare_and_exchange ((void**) &object->qdata, oldvalue,
+                                                     (gpointer) ((gsize) oldvalue | OBJECT_FLOATING_FLAG)));
+      return (gsize) oldvalue & OBJECT_FLOATING_FLAG;
+    case -1:    /* sink if possible */
+      do
+        oldvalue = g_atomic_pointer_get (&object->qdata);
+      while (!g_atomic_pointer_compare_and_exchange ((void**) &object->qdata, oldvalue,
+                                                     (gpointer) ((gsize) oldvalue & ~(gsize) OBJECT_FLOATING_FLAG)));
+      return (gsize) oldvalue & OBJECT_FLOATING_FLAG;
+    default:    /* check floating */
+      return 0 != ((gsize) g_atomic_pointer_get (&object->qdata) & OBJECT_FLOATING_FLAG);
+    }
+}
+
 gboolean
 g_object_is_floating (gpointer _object)
 {
   GObject *object = _object;
   g_return_val_if_fail (G_IS_OBJECT (object), FALSE);
-  return ((gsize) g_atomic_pointer_get (&object->qdata) & OBJECT_FLOATING_FLAG) != 0;
+  return floating_flag_handler (object, 0);
 }
 
 gpointer
 g_object_ref_sink (gpointer _object)
 {
   GObject *object = _object;
-  gpointer oldvalue;
+  gboolean was_floating;
   g_return_val_if_fail (G_IS_OBJECT (object), object);
   g_return_val_if_fail (object->ref_count >= 1, object);
   g_object_ref (object);
-  do
-    {
-      oldvalue = g_atomic_pointer_get (&object->qdata);
-    }
-  while (!g_atomic_pointer_compare_and_exchange ((void**) &object->qdata, oldvalue,
-                                                 (gpointer) ((gsize) oldvalue & ~(gsize) OBJECT_FLOATING_FLAG)));
-  if ((gsize) oldvalue & OBJECT_FLOATING_FLAG)
+  was_floating = floating_flag_handler (object, -1);
+  if (was_floating)
     g_object_unref (object);
   return object;
 }
@@ -1550,15 +1594,11 @@ g_object_ref_sink (gpointer _object)
 void
 g_object_force_floating (GObject *object)
 {
-  gpointer oldvalue;
+  gboolean was_floating;
   g_return_if_fail (G_IS_OBJECT (object));
   g_return_if_fail (object->ref_count >= 1);
-  do
-    {
-      oldvalue = g_atomic_pointer_get (&object->qdata);
-    }
-  while (!g_atomic_pointer_compare_and_exchange ((void**) &object->qdata, oldvalue,
-                                                 (gpointer) ((gsize) oldvalue | OBJECT_FLOATING_FLAG)));
+
+  was_floating = floating_flag_handler (object, +1);
 }
 
 typedef struct {
@@ -2181,5 +2221,39 @@ g_cclosure_new_object_swap (GCallback callback_func,
   return closure;
 }
 
+gsize
+g_object_compat_control (gsize           what,
+                         gpointer        data)
+{
+  switch (what)
+    {
+      gpointer *pp;
+    case 1:     /* floating base type */
+      return G_TYPE_INITIALLY_UNOWNED;
+    case 2:     /* FIXME: remove this once GLib/Gtk+ break ABI again */
+      floating_flag_handler = (guint(*)(GObject*,gint)) data;
+      return 1;
+    case 3:     /* FIXME: remove this once GLib/Gtk+ break ABI again */
+      pp = data;
+      *pp = floating_flag_handler;
+      return 1;
+    default:
+      return 0;
+    }
+}
+
+G_DEFINE_TYPE (GInitiallyUnowned, g_initially_unowned, G_TYPE_OBJECT);
+
+static void
+g_initially_unowned_init (GInitiallyUnowned *object)
+{
+  g_object_force_floating (object);
+}
+
+static void
+g_initially_unowned_class_init (GInitiallyUnownedClass *klass)
+{
+}
+
 #define __G_OBJECT_C__
 #include "gobjectaliasdef.c"
